package com.jdbernard.wdiwtlt.db

import java.lang.reflect.Modifier
import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import javax.sql.DataSource

import groovy.sql.Sql
import groovy.transform.CompileStatic
import groovy.transform.TailRecursive

import org.slf4j.Logger
import org.slf4j.LoggerFactory

import com.jdbernard.wdiwtlt.db.models.*

public class ORM {

    private DataSource dataSource
    private Sql sql

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)
    private static Logger logger = LoggerFactory.getLogger(ORM)

    public ORM(DataSource dataSource) {
        this.dataSource = dataSource
        this.sql = new Sql(dataSource) }

    public void shutdown() { dataSource.shutdown() }

    /// ### Common
    public def getById(int id, Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE id = ?")
            .toString()

        logger.debug("Selecting model.\n\tSQL: {}\n\tPARAMS: {}", query, id)
        return recordToModel(sql.firstRow(query, [id]), modelClass) }

    public def getByName(String name, Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE name = ?")
            .toString()

        logger.debug("Selecting model.\n\tSQL: {}\n\tPARAMS: {}", query, name)
        return recordToModel(sql.firstRow(query, [name]), modelClass) }

    public def getBy(List<String> columns, List<Object> values,
    Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE ")
            .append(columns.collect { it + " = ?" }.join(' AND '))
            .toString()

        logger.debug("Selecting models.\n\tSQL: {}\n\tPARAMS: {}", query, values)
        return sql.rows(query, values)
            .collect { recordToModel(it, modelClass) } }

    public def save(def model) {
        if (model.id > 0) return update(model)
        else return create(model) }

    public def update(def model) {
        def setClauses = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                setClauses << '"' + nameFromModel(field.name) + '"= ?'
                params << field.get(model) }

        def query = new StringBuilder()
            .append("UPDATE ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" SET ")
            .append(setClauses.join(', '))
            .append(" WHERE id = ")
            .append(model.id)
            .toString()

        logger.debug("Updating model.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        return sql.executeUpdate(query, params) }

    public def create(def model) {
        def columns = []
        def params = []

        getInstanceFields(model.class)
            .findAll { it.name != 'id' }
            .each { field ->
                //if (field.class.getAnnotation(Model)) // check to see if we
                                                        // have nested models
                columns << '"' + nameFromModel(field.name) + '"'
                params << field.get(model) }

        def query= new StringBuilder()
            .append("INSERT INTO ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" (")
            .append(columns.join(', '))
            .append(") VALUES (")
            .append((1..columns.size()).collect { '?' }.join(', '))
            .append(")").toString()

        logger.debug("Creating model.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        model.id = sql.executeInsert(query, params)[0][0]
        return 1 }

    public def delete(def model) {
        def query = new StringBuilder()
            .append("DELETE FROM ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append("WHERE id = ?")
            .toString()

        logger.debug("Deleting model.\n\tSQL: {}\n\tPARAMS: {}", query, model.id)
        sql.execute(query, [model.id])
        return sql.updateCount }

    public def associate(String linkTable, Integer firstId, Integer secondId) {
        def query = "INSERT INTO $linkTable VALUES (?, ?) ON CONFLICT DO NOTHING"
        def params = [firstId, secondId]
        logger.debug("Creating association.\n\tSQL: {}\n\tPARAMS: {}",
            query, params)
        return sql.execute(query, params) }

    public def associate(def m1, def m2) {
        return associate(pluralize(nameFromModel(m1.class.simpleName)) +
                   "_" + pluralize(nameFromModel(m2.class.simpleName)),
            m1.id, m2.id) }

    /// ### Album-specific methods
    public Album getAlbumById(int id) { return getById(id, Album) }
    public Album getAlbumByName(String name) { return getByName(name, Album) }
    public Album getAlbumByNameAndArtistId(String name, int artistId) {
        def query = """\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?
            WHERE al.name = ?"""
        def params = [artistId, name]
        logger.debug("Selecting albums.\n\tSQL: {}\n\tPARAMS: {}",
            query, params)
        def albums = sql.rows(query, params)
            .collect { recordToModel(it, Album) }
        return albums ? albums[0] : null }

    public Album getAlbumsByArtistId(int artistId) {
        def query = """\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?"""

        logger.debug("Selecting albums.\n\tSQL: {}\n\tPARAMS: {}", query, artistId)
        return sql.rows(query, [artistId])
            .collect { recordToModel(it, Album) } }

    public List<Album> removeEmptyAlbums() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Artist-specific methods
    public Artist getArtistById(int id) { return getById(id, Artist) }
    public Artist getArtistByName(String name) { return getByName(name, Artist) }
    public Artist getArtistsByAlbum(int albumId) {
        var query = """\
            SELECT ar.*
            FROM artists ar JOIN
                 artists_albums aa ON
                    ar.id = aa.artist_id AND
                    aa.album_id = ?"""
        logger.debug("Selecting artists.\n\tSQL: {}\n\tPARAMS: {}", query, artistId)
        return sql.rows(query, [albumId])
            .collect { recordToModel(it, Artist) } }

    public List<Artist> removeEmptyArtists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    public def addAlbumArtist(int albumId, int artistId) {
        return associate("artists_albums", artistId, albumId) }

    /// ### Bookmark-specific methods
    public Bookmark getBookmarkById(int id) { return getById(id, Bookmark) }
    public Bookmark getBookmarkByName(String name) { return getByName(name, Bookmark) }

    /// ### Image-specific methods
    public Image getImageById(int id) { return getById(id, Image) }

    /// ### MediaFile-specific methods
    public MediaFile getMediaFileById(int id) { return getById(id, MediaFile) }
    public MediaFile getMediaFileByName(String name) { return getByName(name, MediaFile) }

    public MediaFile getMediaFileByFilePath(String filePath) {
        def files = getBy(["file_path"], [filePath], MediaFile)
        return files ? files[0] : null }

    public def associateMediaFileWithAlbum(int mediaFileId, int albumId) {
        return associate("albums_media_files", albumId, mediaFileId) }

    public def associateMediaFileWithArtist(int mediaFileId, int artistId) {
        return associate("artists_media_files", artistId, mediaFileId) }

    public def incrementPlayCount(int mediaFileId) {
        def query = "UPDATE media_files SET play_count = play_count + 1 WHERE ID = ?"
        def params = [mediaFileId]

        logger.debug("Updating media file.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        sql.executeUpdate(query, params)

        query = "SELECT play_count FROM media_files WHERE id = ?"
        logger.debug("Selecting media file play count.\n\tSQL: {}\n\tPARAMS: {}", query, params)
        return sql.firstRow(query, params)[0] }

    public MediaFile incrementPlayCount(MediaFile mf) {
        mf.playCount = incrementPlayCount(mf.id)
        return mf }

    /// ### Playlist-specific methods
    public Playlist getPlaylistById(int id) { return getById(id, Playlist) }
    public Playlist getPlaylistByName(String name) { return getByName(name, Playlist) }

    public List<Playlist> removeEmptyPlaylists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Tag-specific methods
    public Tag getTagById(int id) { return getById(id, Tag) }
    public Tag getTagByName(String name) { return getByName(name, Tag) }

    /// ### Utility functions
    public void withTransaction(Closure c) { sql.withTransaction(c) }
    public static String nameToModel(String name) {
        def pts = name.toLowerCase().split('_')
        return pts.length == 1 ? pts[0] :
            pts[0] + pts[1..-1].collect { it.capitalize() }.join() }

    public static String nameFromModel(String name) {
        return UPPERCASE_PATTERN.matcher(name).
            replaceAll(/$1_$2/).toLowerCase() }

    public static String pluralize(String name) { return name + "s" }

    static def updateModel(def record, def model) {
        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }
        return model }

    static def recordToModel(def record, Class clazz) {
        if (record == null) return null

        def model = clazz.newInstance()

        getInstanceFields(model.class).each { field ->
            field.set(model, record[nameFromModel(field.name)]) }

        return model }

    static def modelToRecord(def model) {
        if (model == null) return null

        def record = [:]

        getInstanceFields(model.class).each { field ->
            record[nameFromModel(field.name)] = field.get(model) }

        return record }

    static def getInstanceFields(Class clazz) {
        return clazz.fields.findAll { !Modifier.isStatic(it.modifiers) } }
}
